import '../models/move_history_model.dart';
import 'package:chessjs/chessjs.dart' as chessjs;
import 'package:chessjs/pieces.dart';
import '../models/chessboard_model.dart';

typedef Null MoveCallback(Map<dynamic, dynamic> moveNotation);
typedef Null CheckMateCallback(chessjs.Color color);
typedef Null CheckCallback(chessjs.Color color);
typedef Null DrawCallback();

class ChessboardController {
  ChessboardController(
      {bool movesEnabled = true,
      bool whiteSideTowardUser = true,
      this.onMove,
      this.onDraw,
      this.onCheck,
      this.onCheckMate,
      String? white,
      bool? enableKeyboard}) {
    this._enableMoves = movesEnabled;
    this._whiteSideTowardUser = whiteSideTowardUser;
    this._white = white ?? 'White';
    this._black = white ?? 'Black';
    this._enableKeyboard = enableKeyboard ?? false;
    buildBoard();
  }

  /// Todo be able to pass a logic, for example a game that already started
  /// and setup the board from a determined position
//  ChessboardController(chessjs.Chess? logic){
//
//    if(logic == null){
//      logic = chessjs.Chess();
//    }
//
//    this.logic = logic;
//
//  }

  /// private fields
  /// model of the board, store the current state of each square
  final ChessboardModel _chessboard = ChessboardModel();

  /// keep track of the move history for the chessboard_notation
  final MoveHistory _moveHistory = MoveHistory();

  /// the actual logic
  final chessjs.Chess _logic = chessjs.Chess();

  /// if board is locked
  late bool _enableMoves;

  /// if keyboard back and forward moves are enable
  late bool _enableKeyboard;

  /// if white is toward the user
  late bool _whiteSideTowardUser;

  /// name to display in white side
  late String _white;

  /// name to display in black side
  late String _black;

  /// callbacks
  MoveCallback? onMove;
  CheckMateCallback? onCheckMate;
  CheckCallback? onCheck;
  DrawCallback? onDraw;

  ChessboardModel get board => _chessboard;
  MoveHistory get moveHistory => _moveHistory;
  chessjs.Chess get logic => _logic;
  bool get movesEnabled => _enableMoves;
  bool get whiteTowardUser => _whiteSideTowardUser;
  String get white => _white;
  String get black => _black;

  /// update white side text
  void setWhite(String white) {
    _white = white;
  }

  /// update black side text
  void setBlack(String black) {
    _black = black;
  }

  /// build the board according to the logic state
  void buildBoard() {
    _chessboard.squares.forEach((square) {
      Piece? piece = _logic.get(square.squareName);
      if (piece != null) {
        square.squareState = piece;
      }
    });
  }

  /// flip board
  void flipBoard() {
    _whiteSideTowardUser = !_whiteSideTowardUser;
  }

  /// get history of moves from logic
  List<String> getSanHistory() {
    List<String> history = _logic.getHistorySAN();
    return history;
  }

  /// Get history of game (moves) with details for each move.
  /// Returns a List<Map>
  List<Map> getVerboseHistory() {
    List<Map> history = _logic.getHistoryVerbose();
    return history;
  }

  /// enable moves in the board
  void enableUserMove() {
    _enableMoves = true;
  }

  /// disable moves in the board
  void disableUserMoves() {
    _enableMoves = false;
  }

  /// enable keyboard
  void enableKeyboard() {
    _enableKeyboard = true;
  }

  /// disable keyboard
  void disableKeyboard() {
    _enableKeyboard = false;
  }

  /// make the actual logic move, used by other functions
  void _makeMove(chessjs.Move move) {
    _logic.makeMove(move);
    _moveHistory.addMoveToHistory(_logic.getHistorySAN().last);
  }

  /// undo a move, this removes the move from the
  /// logic history but not from the [_moveHistory]
  /// In the [_moveHistory] it updates the [_currentIndex]
  void goBack() {
    final undoneMove = _logic.undoMove();
    if (undoneMove != null) {
      _moveHistory.goBackOneMove(undoneMove);
      _updateBoardAfterMove(undoneMove);
    }
  }

  /// go forward and make the last move undone
  void goForward() {
    final chessjs.Move? move = _moveHistory.goForwardOneMove();
    if (move != null) {
      _logic.makeMove(move);
      _updateBoardAfterMove(move);
    }
  }

  /// make move without calling the controller
  /// callbacks. Others make move methods, call
  /// [refreshBoard] at the end, triggering callbacks
  bool makeMoveWithoutCallback(String move) {
    chessjs.Move? moveObj;
    final moves = _logic.generateMoves();
    for (int i = 0; i < moves.length; i++) {
      if (move == _logic.moveToSan(moves[i])) {
        moveObj = moves[i];
        break;
      }
    }
    if (moveObj != null) {
      _makeMove(moveObj);
      _updateBoardAfterMove(moveObj);
      return true;
    } else {
      return false;
    }
  }

  /// returns true if successfully make the move
  /// otherwise false
  bool makeSanMove(String move) {
    chessjs.Move? moveObj;
    final moves = _logic.generateMoves();
    for (int i = 0; i < moves.length; i++) {
      if (move == _logic.moveToSan(moves[i])) {
        moveObj = moves[i];
        break;
      }
    }
    if (moveObj != null) {
      _makeMove(moveObj);
      _updateBoardAfterMove(moveObj);
      refreshBoard();
      return true;
    } else {
      return false;
    }
  }

  /// Attempt to make a move in the board if successfully made the move,
  /// return true and update the respective squares, otherwise return false.
  bool makePrettyMove(Map<String, String> move) {
    final builtMove = _buildPrettyMove(move);
    if (builtMove != null) {
      _makeMove(builtMove);
      _updateBoardAfterMove(builtMove);
      refreshBoard();
      return true;
    } else {
      return false;
    }
  }

  /// build a [Move] from a Map
  chessjs.Move? _buildPrettyMove(Map<String, String> move) {
    final String? promotionStr = move['promotion'];
    final PieceType? promotionPiece;
    switch (promotionStr) {
      case 'n':
        promotionPiece = PieceType.KNIGHT;
        break;
      case 'b':
        promotionPiece = PieceType.BISHOP;
        break;
      case 'r':
        promotionPiece = PieceType.ROOK;
        break;
      case 'q':
        promotionPiece = PieceType.QUEEN;
        break;
      default:
        promotionPiece = null;
    }

    final possibleMoves = _logic.generateMoves();
    chessjs.Move? builtMove;
    possibleMoves.forEach((possibleMove) {
      if (possibleMove.fromAlgebraic == move['from'] &&
          possibleMove.toAlgebraic == move['to'] &&
          possibleMove.promotion == promotionPiece) {
        builtMove = possibleMove;
      }
    });

    if (builtMove == null) {
      return null;
    } else
      return builtMove;
  }

  /// update state of a specific square provided
  /// its squareName and the state from the logic.
  updateSquare(String squareName) {
    _chessboard.updateSquareState(
      squareName,
      _logic.get(squareName),
    );
  }

  /// after playing, update squares involved in the move
  _updateBoardAfterMove(chessjs.Move move) {
    int flags = move.flags;

    /// update square from where the move originated
    updateSquare(move.fromAlgebraic);

    /// if castling, update squares associated
    if ((flags & chessjs.BITS_KSIDE_CASTLE) != 0) {
      if (move.color == chessjs.WHITE) {
        updateSquare('h1');
        updateSquare('g1');
        updateSquare('f1');
      } else {
        updateSquare('h8');
        updateSquare('g8');
        updateSquare('f8');
      }
    } else if ((flags & chessjs.BITS_QSIDE_CASTLE) != 0) {
      if (move.color == chessjs.WHITE) {
        updateSquare('a1');
        updateSquare('c1');
        updateSquare('d1');
      } else {
        updateSquare('a8');
        updateSquare('c8');
        updateSquare('d8');
      }
    }

    /// if en passant, update squares associated
    else if ((flags & (chessjs.BITS_EP_CAPTURE)) != 0) {
      late final squareToUpdate;

      String to = move.toAlgebraic;
      if (move.color == chessjs.WHITE) {
        squareToUpdate = to.replaceAll('6', '5');
      } else if (move.color == chessjs.BLACK) {
        squareToUpdate = to.replaceAll('3', '4');
      }
      updateSquare(squareToUpdate);
      updateSquare(move.fromAlgebraic);
      updateSquare(move.toAlgebraic);
    }

    /// otherwise just update the square where the move finished
    else {
      updateSquare(move.toAlgebraic);
    }
//    notifyListeners();
  }

  /// use [refreshBoard] to force controller callbacks
  /// to run, for example onMove, onCheck...
  void refreshBoard() {
    if (logic.inCheckmate && onCheckMate != null) {
      onCheckMate!(logic.turn == chessjs.WHITE ? chessjs.WHITE : chessjs.BLACK);
    } else if ((logic.inDraw ||
            logic.inStalemate ||
            logic.inThreefoldRepetition ||
            logic.insufficientMaterial) &&
        onDraw != null) {
      onDraw!();
    } else if (logic.inCheck && onCheck != null) {
      onCheck!(logic.turn == chessjs.WHITE ? chessjs.WHITE : chessjs.BLACK);
    } else {
      if (onMove != null) onMove!(logic.getHistoryVerbose().last);
    }
  }
}
