import 'package:chessjs/chessjs.dart' as chessjs;
import 'package:chessjs/pieces.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import '../utils/constants.dart';
import 'controller.dart';
import '../providers/providers.dart';
import '../utils/utils.dart';
import 'package:hooks_riverpod/hooks_riverpod.dart';

typedef Null MoveCallback(Map<String, String> moveNotation);

class BoardSquare extends HookConsumerWidget {
  BoardSquare({
    required Key key,
    required this.squareName,
    required this.size,
    required this.controller,
    required this.onMove,
  });

  final ChessboardController controller;
  final String squareName;
  final double size;
  final MoveCallback onMove;

  @override
  Widget build(BuildContext context, WidgetRef ref) {
    /// get state of square
    Piece? squareState = ref.watch(boardProvider(controller)
        .select((board) => board.squares[conversion[squareName]!].squareState));

    return Expanded(
        flex: 1,
        child: DragTarget(
          builder: (context, accepted, rejected) {
//            /// TODO fix this - for now use this workaround
//            /// there are issues when promoting
            squareState = controller.logic.get(squareName);

            /// get the piece on the square
            return squareState != null
                ?

                /// if there is a piece on the square, returns a draggable image of the piece
                Draggable(
                    child: getImageToDisplay(
                      size: size,
                      squareState: squareState,
                    ),
                    feedback: getImageToDisplay(
                        size: (1.15 * size), squareState: squareState),

                    /// Data to be dropped in this drag target: the squareName where
                    /// the move originated, and the type and color of the piece that was
                    /// on that square
                    data: [
                      squareName,
                      squareState!.type.toUpperCase(),
                      squareState!.color,
                    ],
                  )

                /// Changing here Container for SizedBox wont allow to make a move
                /// Container expands to the child size, SizedBox wont, so the drag target
                /// property will be on a 0 size widget and wont work
                : Container();

            /// accepted information about the move
            /// the square from where the move originated is moveInfo[0],
            /// the type of piece is moveInfo[1], and the color is moveInfo[2]
          },
          onWillAccept: (List? moveInfo) {
            if (moveInfo != null) {
              /// If dragging and dropping in same place return false
              if (moveInfo[0] != squareName) {
                /// moveInfo should have 3 items
                if (moveInfo.length == 3) {
                  /// Accept only if the game turn is right and the enableUserMoves is true
                  if (moveInfo[2] == controller.logic.turn &&
                      controller.movesEnabled) {
                    /// DO NOT ACCEPT ILLEGAL MOVES
                    final legalMoves = controller.logic.generateMoves();
                    legalMoves.removeWhere((move) {
                      /// if [from] move part is different, remove item
                      if (move.fromAlgebraic != moveInfo[0]) {
                        return true;
                      } else {
                        /// if [to] move part is different, also remove item
                        if (move.toAlgebraic != squareName) {
                          return true;
                        } else {
                          return false;
                        }
                      }
                    });

                    /// If there were not matches of intended move don't accept
                    if (legalMoves.isEmpty) {
                      return false;
                    } else {
                      return true;
                    }
                  } else {
                    return false;
                  }
                } else {
                  return false;
                }
              } else
                return false;
            } else
              return false;
          },
          onAccept: (List moveInfo) async {
            Map<String, String>? move;

            /// handle promotion
            if (moveInfo[1] == "P" &&
                ((moveInfo[0][1] == "7" &&
                        squareName[1] == "8" &&
                        moveInfo[2] == chessjs.WHITE) ||
                    (moveInfo[0][1] == "2" &&
                        squareName[1] == "1" &&
                        moveInfo[2] == chessjs.BLACK))) {
              /// get the promotion piece selected by the user from the dialog
              String? val =
                  await promotionDialog(context, controller.logic.turn);
              if (val == 'n' || val == 'b' || val == 'r' || val == 'q') {
                /// move to return in case of promotion
                move = {
                  "from": moveInfo[0],
                  "to": squareName,
                  "promotion": val!,
                };
                onMove(move);
              }
            } else {
              move = {"from": moveInfo[0], "to": squareName};
              onMove(move);
            }
          },
        ));
  }
}
